@@ -1,13 +1,14 @@
 /*
  * logger.c
  *
- * Copyright (c) 2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfo.h"
 
 #include <dirent.h>
 #include <fcntl.h>
+#include <sys/stat.h>
 #include <time.h>
 
 #ifndef WIN32
@@ -15,6 +16,7 @@
 #endif
 
 static time_t		logger_reload_time = 0;
+static List		   *adjust_log_list = NIL;
 
 /*---- GUC variables (logger) ----------*/
 static char		   *my_log_directory;
@@ -27,6 +29,13 @@ static char		   *my_textlog_filename;
 static char		   *my_textlog_line_prefix;
 static int			my_textlog_min_messages;
 static int			my_textlog_permission;
+static bool			my_adjust_log_level;
+static char		   *my_adjust_log_info;
+static char		   *my_adjust_log_notice;
+static char		   *my_adjust_log_warning;
+static char		   *my_adjust_log_error;
+static char		   *my_adjust_log_log;
+static char		   *my_adjust_log_fatal;
 /*-----------------------*/
 
 #if PG_VERSION_NUM < 90000
@@ -65,6 +74,12 @@ typedef struct Logger
 	size_t	fields[CSV_COLS + 1];	/* field offsets in buffer */
 } Logger;
 
+typedef struct AdjustLog
+{
+	char	*sqlstate;	/* SQL STATE to be adjusted */
+	int		 elevel;	/* log level */
+} AdjustLog;
+
 static List			   *log_queue = NIL;
 static pthread_mutex_t	log_lock;
 static int				recursive_level = 0;
@@ -80,9 +95,12 @@ static void init_log(Log *log, const char *buf, size_t len, const size_t fields[
 static void logger_close(Logger *logger);
 static bool logger_next(Logger *logger, const char *pg_log);
 static void get_csvlog(char csvlog[], const char *prev, const char *pg_log);
+static void adjust_log(Log *log);
 static void replace_extension(char path[], const char *extension);
 static void assign_textlog_path(Logger *logger, const char *pg_log);
 static void assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog);
+static List *add_adlog(List *adlog_list, int elevel, char *rawstring);
+static char *b_trim(char *str);
 
 void
 logger_init(void)
@@ -365,6 +383,33 @@ reload_params(void)
 	free(my_syslog_ident);
 	my_syslog_ident = pgut_strdup(syslog_ident);
 
+	/* adjust log level */
+	my_adjust_log_level = adjust_log_level;
+	if (my_adjust_log_level)
+	{
+		free(my_adjust_log_info);
+		my_adjust_log_info = pgut_strdup(adjust_log_info);
+		free(my_adjust_log_notice);
+		my_adjust_log_notice = pgut_strdup(adjust_log_notice);
+		free(my_adjust_log_warning);
+		my_adjust_log_warning = pgut_strdup(adjust_log_warning);
+		free(my_adjust_log_error);
+		my_adjust_log_error = pgut_strdup(adjust_log_error);
+		free(my_adjust_log_log);
+		my_adjust_log_log = pgut_strdup(adjust_log_log);
+		free(my_adjust_log_fatal);
+		my_adjust_log_fatal = pgut_strdup(adjust_log_fatal);
+
+		list_destroy(adjust_log_list, free);
+		adjust_log_list = NIL;
+		adjust_log_list = add_adlog(adjust_log_list, FATAL, my_adjust_log_fatal);
+		adjust_log_list = add_adlog(adjust_log_list, LOG, my_adjust_log_log);
+		adjust_log_list = add_adlog(adjust_log_list, ERROR, my_adjust_log_error);
+		adjust_log_list = add_adlog(adjust_log_list, WARNING, my_adjust_log_warning);
+		adjust_log_list = add_adlog(adjust_log_list, NOTICE, my_adjust_log_notice);
+		adjust_log_list = add_adlog(adjust_log_list, INFO, my_adjust_log_info);
+	}
+
 	pgut_log_level = Min(textlog_min_messages, syslog_min_messages);
 
 	pthread_mutex_unlock(&reload_lock);
@@ -481,6 +526,7 @@ logger_parse(Logger *logger, const char *pg_log)
 	while (logger_next(logger, pg_log))
 	{
 		Log		log;
+		int		save_elevel;
 
 		init_log(&log, logger->buf.data, logger->buf.len, logger->fields);
 
@@ -505,6 +551,16 @@ logger_parse(Logger *logger, const char *pg_log)
 				continue;
 			}
 
+			/* maintenance requested ? */
+			if (strcmp(log.message, LOGMSG_MAINTENANCE) == 0)
+			{
+				pthread_mutex_lock(&reload_lock);
+				free((char *) maintenance_requested);
+				maintenance_requested = pgut_strdup(log.detail);
+				pthread_mutex_unlock(&reload_lock);
+				continue;
+			}
+
 			/* restart requested ? */
 			if (strcmp(log.message, LOGMSG_RESTART) == 0)
 			{
@@ -525,10 +581,16 @@ logger_parse(Logger *logger, const char *pg_log)
 #endif
 		}
 
-		/* route the log to syslog and textlog. */
+		/*
+		 * route the log to syslog and textlog.
+		 * if log level adjust enabled, adjust the log level.
+		 */
+		save_elevel = log.elevel;
+		if (my_adjust_log_level)
+			adjust_log(&log);
 		logger_route(logger, &log);
 
-		if (log.elevel == LOG)
+		if (save_elevel == LOG)
 		{
 			/* setting parameters reloaded ? */
 			if (strcmp(log.message, msg_sighup) == 0)
@@ -574,7 +636,11 @@ logger_close(Logger *logger)
 		fclose(logger->textlog);
 		logger->textlog = NULL;
 
-		/* overwrite existing .log file; it must be empty. */
+		/* overwrite existing .log file; it must be empty.
+		 * (Note : Some error messages through system() (eg. recovery_command)
+		 *  outputs to the .log file  from postgres's logger, so sometimes
+		 *  .log file will not be empty. At the moment we overwrite without check.
+		 */
 		if (logger->csv_path[0] && stat(logger->csv_path, &st) == 0)
 		{
 			char	path[MAXPGPATH];
@@ -617,10 +683,15 @@ logger_next(Logger *logger, const char *pg_log)
 
 		/*
 		 * csvlog files that have empty *.log have not been parsed yet
-		 * because ostgres logger make an empty log file.
+		 * because postgres logger make an empty log file.
+		 * (Note: Some "cannot stat" error messages are output to *.log,
+		 *  so we check logger->fp and csvlog again.
 		 */
-		if (stat(textlog, &st) == 0 && st.st_size > 0)
-			return false;	/* already parsed log */
+		 if (stat(textlog, &st) == 0 && st.st_size > 0)
+		 {
+			if (logger->fp && strcmp(logger->csv_name, csvlog) == 0)
+				return false;	/* already parsed log */
+		 }
 
 		logger_close(logger);
 		assign_textlog_path(logger, pg_log);
@@ -702,6 +773,24 @@ get_csvlog(char csvlog[], const char *prev, const char *pg_log)
 	closedir(dir);
 }
 
+static void
+adjust_log(Log *log)
+{
+	ListCell	*cell;
+
+	foreach(cell, adjust_log_list)
+	{
+		AdjustLog *adlog = (AdjustLog *) lfirst(cell);
+
+		if (strcmp(adlog->sqlstate, log->sqlstate) == 0)
+		{
+			log->elevel = adlog->elevel;
+			elog(DEBUG2, "adjust log level -> %d: sqlstate=\"%s\"", log->elevel, log->sqlstate);
+			break;
+		}
+	}
+}
+
 static void
 replace_extension(char path[], const char *extension)
 {
@@ -729,3 +818,45 @@ assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog)
 	logger->csv_name = logger->csv_path + strlen(pg_log) + 1;
 	logger->csv_offset = 0;
 }
+
+static List *
+add_adlog(List *adlog_list, int elevel, char *rawstring)
+{
+	char	*token;
+
+	token = strtok(rawstring, ",");
+	while (token)
+	{
+		AdjustLog *adlog = pgut_malloc(sizeof(AdjustLog));;
+		adlog->elevel = elevel;
+		adlog->sqlstate = b_trim(token);
+		adlog_list = lappend(adlog_list, adlog);
+		token = strtok(NULL, ",");
+	}
+	return adlog_list;
+}
+
+/*
+ * Note: this function modify the argument string
+ */
+static char *
+b_trim(char *str)
+{
+	size_t	 len;
+	char	*start;
+
+	if (str == NULL)
+		return NULL;
+
+	/* remove space character from prefix */
+	len = strlen(str);
+	while (len > 0 && isspace(str[len - 1])) { len--; }
+	str[len] = '\0';
+
+	/* remove space character from suffix */
+	start = str;
+	while (isspace(start[0])) { start++; }
+	memmove(str, start, strlen(start) + 1);
+
+	return str;
+}
