@@ -1,7 +1,7 @@
 /*
  * pg_statsinfo.c
  *
- * Copyright (c) 2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfo.h"
@@ -10,7 +10,7 @@
 
 #include "miscadmin.h"
 
-const char *PROGRAM_VERSION	= "2.0.0";
+const char *PROGRAM_VERSION	= "2.1.0";
 const char *PROGRAM_URL		= "http://pgstatsinfo.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 
@@ -47,8 +47,18 @@ char		   *textlog_filename;
 char		   *textlog_line_prefix;
 int				textlog_min_messages;
 int				textlog_permission;
+bool			adjust_log_level;
+char		   *adjust_log_info;
+char		   *adjust_log_notice;
+char		   *adjust_log_warning;
+char		   *adjust_log_error;
+char		   *adjust_log_log;
+char		   *adjust_log_fatal;
 /*---- GUC variables (writer) ----------*/
 char		   *repository_server;
+bool		    enable_maintenance;
+time_t			maintenance_time;
+int				repository_keepday;
 /*---- message format ----*/
 char		   *msg_debug;
 char		   *msg_info;
@@ -86,7 +96,11 @@ static bool assign_int(const char *value, void *var);
 static bool assign_elevel(const char *value, void *var);
 static bool assign_syslog(const char *value, void *var);
 static bool assign_string(const char *value, void *var);
+static bool assign_bool(const char *value, void *var);
+static bool assign_time(const char *value, void *var);
 static void after_readopt(void);
+static bool decode_time(const char *field, int *hour, int *min, int *sec);
+static int strtoi(const char *nptr, char **endptr, int base);
 
 /* parameters */
 static struct ParamMap PARAM_MAP[] =
@@ -115,6 +129,16 @@ static struct ParamMap PARAM_MAP[] =
 	{GUC_PREFIX ".textlog_filename", assign_string, &textlog_filename},
 	{GUC_PREFIX ".textlog_line_prefix", assign_string, &textlog_line_prefix},
 	{GUC_PREFIX ".textlog_permission", assign_int, &textlog_permission},
+	{GUC_PREFIX ".adjust_log_level", assign_bool, &adjust_log_level},
+	{GUC_PREFIX ".adjust_log_info", assign_string, &adjust_log_info},
+	{GUC_PREFIX ".adjust_log_notice", assign_string, &adjust_log_notice},
+	{GUC_PREFIX ".adjust_log_warning", assign_string, &adjust_log_warning},
+	{GUC_PREFIX ".adjust_log_error", assign_string, &adjust_log_error},
+	{GUC_PREFIX ".adjust_log_log", assign_string, &adjust_log_log},
+	{GUC_PREFIX ".adjust_log_fatal", assign_string, &adjust_log_fatal},
+	{GUC_PREFIX ".enable_maintenance", assign_bool, &enable_maintenance},
+	{GUC_PREFIX ".maintenance_time", assign_time, &maintenance_time},
+	{GUC_PREFIX ".repository_keepday", assign_int, &repository_keepday},
 	{":debug", assign_string, &msg_debug},
 	{":info", assign_string, &msg_info},
 	{":notice", assign_string, &msg_notice},
@@ -397,12 +421,51 @@ ensure_schema(PGconn *conn, const char *schema)
 	res = pgut_execute(conn,
 			"SELECT nspname FROM pg_namespace WHERE nspname = $1",
 			1, &schema);
+	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	{
+		PQclear(res);
+		return false;
+	}
 	ok = (PQntuples(res) > 0);
 	PQclear(res);
 	/* TODO: check installed schema version */
 	if (ok)
 		return true;	/* ok, installed */
 
+	/* iff schema is "statsrepo", check repository server version */
+	if (strcmp(schema, "statsrepo") == 0)
+	{
+		int		server_version;
+		bool	installed;
+
+		server_version = get_server_version(conn);
+
+		if (server_version < 0)
+			return false;
+		else if (server_version >= 80400)
+		{
+			/* sets repository schema for partitioning */
+			schema = "statsrepo_partition";
+
+			/* create language 'PL/pgSQL' */
+			res = pgut_execute(conn,
+				"SELECT 1 FROM pg_language WHERE lanname = 'plpgsql'", 0, NULL);
+			if (PQresultStatus(res) != PGRES_TUPLES_OK)
+			{
+				PQclear(res);
+				return false;
+			}
+			installed = (PQntuples(res) > 0);
+			PQclear(res);
+			if (!installed)
+			{
+				if (pgut_command(conn,
+					"CREATE LANGUAGE plpgsql", 0, NULL) != PGRES_COMMAND_OK)
+					return false;
+			}
+		}
+	}
+
 	/* read $PGSHARE/contrib/pg_{schema}.sql into buffer. */
 	snprintf(path, MAXPGPATH, "%s/contrib/pg_%s.sql", share_path, schema);
 	if ((fp = pgut_fopen(path, "rt")) == NULL)
@@ -456,6 +519,32 @@ assign_int(const char *value, void *var)
 	return parse_int32(value, (int32 *) var);
 }
 
+static bool
+assign_bool(const char *value, void *var)
+{
+	return parse_bool(value, var);
+}
+
+static bool
+assign_time(const char *value, void *var)
+{
+	struct tm	*tm;
+	time_t		 now;
+	int			 hour, min, sec;
+
+	if (!decode_time(value, &hour, &min, &sec))
+		return false;
+
+	now = time(NULL);
+	tm = localtime(&now);
+	tm->tm_hour = hour;
+	tm->tm_min = min;
+	tm->tm_sec = sec;
+
+	*(time_t *)var = mktime(tm);
+	return true;
+}
+
 static bool
 assign_elevel(const char *value, void *var)
 {
@@ -598,6 +687,10 @@ after_readopt(void)
 		if (flags)
 			checkpoint_starting_prefix_len = flags - msg_checkpoint_starting;
 	}
+
+	if (!enable_maintenance)
+		elog(NOTICE,
+			"automatic maintenance is disable. Please note the data size of the repository");
 }
 
 /*
@@ -609,3 +702,130 @@ delay(void)
 	if (shutdown_state < SHUTDOWN_REQUESTED)
 		sleep(1);
 }
+
+/*
+ * get local timestamp by character string
+ * return format : "YYYY-MM-DD HH:MM:SS.FFFFFF"
+ */
+char *
+getlocaltimestamp(void)
+{
+#ifndef WIN32
+	struct timeval	 tv;
+	struct tm		*ts;
+#else
+	SYSTEMTIME		 stTime;
+#endif
+	char			*tp;
+
+	if ((tp = (char *)pgut_malloc((size_t)32)) == NULL)
+		return NULL;
+
+	memset(tp, 0x00, 32);
+
+#ifndef WIN32
+	if (gettimeofday(&tv, NULL) != 0)
+	{
+		free(tp);
+		ereport(ERROR,
+			(errcode_errno(),
+			 errmsg("gettimeofday function call failed")));
+		return NULL;
+	}
+
+	if ((ts = localtime(&tv.tv_sec)) == NULL)
+	{
+		free(tp);
+		ereport(ERROR,
+			(errcode_errno(),
+			 errmsg("localtime function call failed")));
+		return NULL;
+	}
+
+	snprintf(tp, 32, "%04d-%02d-%02d %02d:%02d:%02d.%ld",
+			ts->tm_year + 1900,
+			ts->tm_mon + 1,
+			ts->tm_mday,
+			ts->tm_hour,
+			ts->tm_min,
+			ts->tm_sec,
+			tv.tv_usec);
+#else
+	GetLocalTime(&stTime);
+
+	snprintf(tp, 32, "%04d-%02d-%02d %02d:%02d:%02d.%ld",
+			stTime.wYear,
+			stTime.wMonth,
+			stTime.wDay,
+			stTime.wHour,
+			stTime.wMinute,
+			stTime.wSecond,
+			stTime.wMilliseconds);
+
+#endif   /* WIN32 */
+
+	return tp;
+}
+
+int
+get_server_version(PGconn *conn)
+{
+	PGresult	*res;
+	int			 server_version_num;
+
+	res = pgut_execute(conn, "SHOW server_version_num", 0, NULL);
+	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+		server_version_num = -1;
+	else
+		server_version_num = atoi(PQgetvalue(res, 0, 0));
+	PQclear(res);
+
+	return server_version_num;
+}
+
+static bool
+decode_time(const char *field, int *hour, int *min, int *sec)
+{
+	char	*cp;
+
+	errno = 0;
+	*hour = strtoi(field, &cp, 10);
+	if (errno == ERANGE || *cp != ':')
+		return false;
+	errno = 0;
+	*min = strtoi(cp + 1, &cp, 10);
+	if (errno == ERANGE)
+		return false;
+	if (*cp == '\0')
+		*sec = 0;
+	else if (*cp == ':')
+	{
+		errno = 0;
+		*sec = strtoi(cp + 1, &cp, 10);
+		if (errno == ERANGE || *cp != '\0')
+			return false;
+	}
+	else
+		return false;
+
+	/* sanity check */
+	if (*hour < 0 || *hour > 23 ||
+		*min < 0 || *min > 59 ||
+		*sec < 0 || *sec > 59)
+		return false;
+
+	return true;
+}
+
+static int
+strtoi(const char *nptr, char **endptr, int base)
+{
+	long		val;
+
+	val = strtol(nptr, endptr, base);
+#ifdef HAVE_LONG_INT_64
+	if (val != (long) ((int32) val))
+		errno = ERANGE;
+#endif
+	return (int) val;
+}
