@@ -2,7 +2,7 @@
  *
  * pg_statsinfo.h
  *
- * Copyright (c) 2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  *
  *-------------------------------------------------------------------------
  */
@@ -19,9 +19,10 @@
 
 ///#define USE_DAEMON				/* become daemon? */
 
-#define DB_MAX_RETRY		10	/* max retry count for database operations */
-#define LOGTIME_LEN			40	/* buffer size for timestamp */
-#define LOGCODE_LEN			6	/* buffer size for sqlcode */
+#define DB_MAX_RETRY		10		/* max retry count for database operations */
+#define LOGTIME_LEN			40		/* buffer size for timestamp */
+#define LOGCODE_LEN			6		/* buffer size for sqlcode */
+#define SECS_PER_DAY		86400	/* seconds per day */
 
 /* shutdown state */
 typedef enum ShutdownState
@@ -57,8 +58,18 @@ extern char		   *textlog_filename;
 extern char		   *textlog_line_prefix;
 extern int			textlog_min_messages;
 extern int			textlog_permission;
+extern bool			adjust_log_level;
+extern char		   *adjust_log_info;
+extern char		   *adjust_log_notice;
+extern char		   *adjust_log_warning;
+extern char		   *adjust_log_error;
+extern char		   *adjust_log_log;
+extern char		   *adjust_log_fatal;
 /*---- GUC variables (writer) ----------*/
 extern char		   *repository_server;
+extern bool		    enable_maintenance;
+extern time_t		maintenance_time;
+extern int			repository_keepday;
 /*---- message format ----*/
 extern char		   *msg_debug;
 extern char		   *msg_info;
@@ -90,9 +101,11 @@ extern pthread_t	th_writer;
 
 /* collector.c */
 extern pthread_mutex_t	reload_lock;
+extern pthread_mutex_t	maintenance_lock;
 extern volatile time_t	server_reload_time;
 extern volatile time_t	collector_reload_time;
 extern volatile char   *snapshot_requested;
+extern volatile char   *maintenance_requested;
 
 /* queue item for writer */
 typedef struct QueueItem	QueueItem;
@@ -159,11 +172,14 @@ extern bool write_textlog(const Log *log, const char *prefix,
 extern bool parse_checkpoint(const char *message, const char *timestamp);
 /* autovacuum.c */
 extern bool parse_autovacuum(const char *message, const char *timestamp);
+/* maintenance.c */
+extern bool check_maintenance(const char *message);
 
 /* writer.c */
 extern void writer_init(void);
 extern void *writer_main(void *arg);
 extern void writer_send(QueueItem *item);
+extern void do_maintenance(time_t repository_keepday);
 
 /* pg_statsinfo.c */
 extern bool postmaster_is_alive(void);
@@ -172,5 +188,7 @@ extern int str_to_elevel(const char *value);
 extern const char *elevel_to_str(int elevel);
 extern void shutdown_progress(ShutdownState state);
 extern void delay(void);
+extern char *getlocaltimestamp(void);
+extern int get_server_version(PGconn *conn);
 
 #endif   /* PG_STATSINFO_H */
